# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide cmdbar 1.2

namespace eval cmdbar {
  variable options
  array set options {
    Pathname  {}
    Binding {}
    HistoryMaxSize -1
    Debug 0
  }
  variable lock 0
  variable Debug 1
  variable history [list]
  variable execFlags [list]
}

proc cmdbar::init {w {binding {. <colon>}}} {
  variable options
  entry $w -state disabled -disabledforeground black
  bind [lindex $binding 0] [lindex $binding 1] [list apply {{w} {
      focus -force $w
      ::cmdbar::enterCmdBar
    }} $w]
  set options(Binding) $binding
  set options(Pathname) $w
}

proc cmdbar::enterCmdBar {{text :}} {
  variable options
  if {$::cmdbar::lock} {return}
  set ::cmdbar::lock 1
  set ::cmdbar::bindStore [::cmdbar::deleteBindings]
  $options(Pathname) configure -state normal
  $options(Pathname) delete 0 end
  $options(Pathname) insert 0 $text
  focus $options(Pathname)
  grab set $options(Pathname)
  bind $options(Pathname) <Return> [list apply {{path} {
      ::cmdbar::execCmdBar [$path get] 
    }} $options(Pathname)]
  bind $options(Pathname) <Escape> [list apply {{} {
      ::cmdbar::releaseCmdBar
    }}]
}

proc cmdbar::writeToCmdBar {txt args} {
  variable options
  # do not write to cmdbar if it's occupied, but it may be forced
  if {[set f [lsearch $args {-force}]] >= 0 && [lindex $args $f+1]} {set forced 1} else {set forced 0}
  if {$::cmdbar::lock && !$forced} {return}
  # do not allow enter the commandline while messaging
  set ::cmdbar::lock 1
  set delay 1000
  $options(Pathname) configure -state normal
  foreach {opt value} $args {
    switch $opt {
      -fg {$options(Pathname) configure -fg $value}
      -bg {$options(Pathname) configure -readonlybackground $value}
      -delay {set delay $value}
      default {}
    }
  }
  $options(Pathname) delete 0 end
  if {$txt ne {}} {
    $options(Pathname) insert 0 $txt
    if {$delay ne {forever}} {
      # set color back to gray after some delay
      after $delay $options(Pathname) configure -readonlybackground grey
    }
    # allow entering the command line after 1 second. This is not affected by the -delay parameter
    after 1000 [list apply {{w lockvar} {
        upvar $lockvar mylock
        set mylock 0
      }} $options(Pathname) ::cmdbar::lock]
  } else {
    $options(Pathname) configure -readonlybackground grey
    set ::cmdbar::lock 1
  }

  $options(Pathname) configure -state readonly
}

# Execute the command on the command line
proc cmdbar::execCmdBar {cmd} {
  variable options
  variable execFlags
  if {$cmd eq {}} {
    set cmd [$options(Pathname) get]
  }
  ::cmdbar::releaseCmdBar
  # $cmd must be a valid command: it must start with a colon, but may
  # not contain sequences of 2 or more colons
  if {[regexp {^(?!\:)|\:{2,}} $cmd]} {
    bell
    ::cmdbar::writeToCmdBar "Invalid command: $cmd" -bg red
  } else {
    switch -- [catch {eval $cmd} result erropts] {
      1 {
        # error
        bell
        set msg "Command $cmd failed: $result"
        if {$options(Debug)} {
          ::cmdbar::releaseCmdBar
          return -options $erropts $msg
        } else {
          ::cmdbar::writeToCmdBar $msg -bg red
        }
        if {$::cmdbar::Debug} {error $result $::errorInfo}
      }
      3 -
      4 {
        # warning
        set msg "$cmd -> Warning: $result"
        ::cmdbar::writeToCmdBar $msg -bg yellow
      }
      0 -
      default {
        # ok
        set msg "$cmd -> $result"
        ::cmdbar::writeToCmdBar $msg -bg green
      }
    }
  }
  set disable 1
  set addHistory 1
  foreach flag $execFlags {
    switch -- $flag {
      -nodisable {set disable 0}
      -nohistory {set addHistory 0}
    }
  }
  set execFlags [list]
  if {$addHistory} {
    # add command to history
    ::cmdbar::historyAdd $cmd
  }
  if {$disable} {
    after 1000 $options(Pathname) configure -state disabled -bg white -fg black
  }
}

proc cmdbar::deleteBindings {} {
  foreach win [join [list . [winfo children .]]] {
    foreach seq [bind $win] {
      lappend b $win $seq [bind $win $seq]
      bind $win $seq {}
    }
  }
  return $b
}

proc cmdbar::restoreBindings {b} {
  foreach {win seq cmd} $b {
    bind $win $seq $cmd
  }
}

proc cmdbar::releaseCmdBar {} {
  variable options
  focus [focus -lastfor $options(Pathname)]
  grab release $options(Pathname)
  set ::cmdbar::lock 0
  catch {::cmdbar::restoreBindings $::cmdbar::bindStore}
  set ::cmdbar::bindstore {}
  bind $options(Pathname) <Return> {}
  bind $options(Pathname) <Escape> {}
}

# add command to the history
proc cmdbar::historyAdd {{cmd {}}} {
  variable options
  variable history
  if {$options(HistoryMaxSize) < 0} {
    lappend history $cmd
  } elseif {$options(HistoryMaxSize) > 0} {
    if {[llength $history] < $options(HistoryMaxSize)} {
      lappend history $cmd
    } else {
      set history [concat [lrange $history 1 end] $cmd]
    }
  } 
  # elseif $options(HistoryMaxSize) == 0 => no history
} 

# get a command from the history
proc cmdbar::historyGet {{T 0}} {
  variable options
  variable history
  if {$T < 0} {
    return [list [llength $history] $options(HistoryMaxSize)]
  } elseif {$T >= [llength $history]} {
    # this also applies to llength $history == 0
    return {}
  } else {
    return [lindex $history end-$T]
  }
}

proc cmdbar::addExecFlags {{flags {}}} {
  variable execFlags
  if {$flags eq {}} {
    # clear flags
    set execFlags [list]
  } else {
    set execFlags [concat $execFlags $flags]
  }
}

